package edg;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import edg.ASTBuilder.Where;
import edg.graph.EDG;
import edg.graph.NodeInfo;

public abstract class EDGFactory
{
	private EDG edg;
	private final Stack<Branch> branches = new Stack<Branch>();

	private <R> void processElements(R[] elements)
	{
		for (int elementIndex = 0; elementIndex < elements.length; elementIndex++)
		{
			final R element = elements[elementIndex];
			this.processElement(element, elementIndex + 1, elements.length);
		}
	}
	private void processElement(Object element, int index, int length)
	{
		final Map<String, Object> info = new HashMap<String, Object>();
		final Branch parent = this.branches.isEmpty() ? null : this.branches.get(this.branches.size() - 1);
		if (parent != null)
		{
			parent.setIndex(index);
			parent.setLength(length);
		}

		final boolean isPatternZone = this.isPatternZone(this.branches);
		info.put("ancestors", this.branches);
		info.put("parent", parent);
		info.put("patternZone", isPatternZone);

		this.processElement(element, info);
	}
	protected abstract void processElement(Object element, Map<String, Object> info);

	protected <R> EDG createEDG(Iterable<R> classes, LDASTNodeInfo info)
	{
		final R[] classes0 = this.getArray(classes);
		return this.createEDG(true, classes0, info);
	}
	protected <R> EDG createEDG(boolean generateArcs, Iterable<R> classes, LDASTNodeInfo info)
	{
		final R[] classes0 = this.getArray(classes);
		return this.createEDG(generateArcs, classes0, info);
	}
	protected <R> EDG createEDG(R[] classes, LDASTNodeInfo info)
	{
		return this.createEDG(true, classes, info);
	}
	protected <R> EDG createEDG(boolean generateArcs, R[] classes, LDASTNodeInfo info)
	{
		this.branches.clear();

		this.edg = ASTBuilder.createEDG(info);
		this.processElements(classes);
		ASTBuilder.completeEDG(this.edg);

		if (generateArcs)
			ASTBuilder.generateDependencies(this.edg);

		return this.edg;
	}

	protected <R> void addModule(String name, Iterable<R> members, LDASTNodeInfo info)
	{
		final R[] members0 = this.getArray(members);
		this.addModule(name, members0, info);
	}
	protected <R> void addModule(String name, R[] members, LDASTNodeInfo info)
	{
		final int moduleId = ASTBuilder.addModule(this.edg, name, info);

		this.branches.push(new Branch(moduleId, NodeInfo.Type.Module, info));
		this.processElements(members);
		this.branches.pop();
	}

	protected <R> void addRoutine(String name, Iterable<R> clauses, LDASTNodeInfo info)
	{
		final R[] clauses0 = this.getArray(clauses);
		this.addRoutine(name, clauses0, info);
	}
	protected <R> void addRoutine(String name, R[] clauses, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final Where where = parent.getWhere();
		final int routineId = ASTBuilder.addRoutine(this.edg, parentId, where, name, info);

		this.branches.push(new Branch(routineId, NodeInfo.Type.Routine, info));
		this.processElements(clauses);
		this.branches.pop();
	}

	protected <R, S, T> void addClause(Iterable<R> parameters, S guard, T body, LDASTNodeInfo info)
	{
		final R[] parameters0 = this.getArray(parameters);
		this.addClause(parameters0, guard, body, info);
	}
	protected <R, S, T> void addClause(R[] parameters, S guard, T body, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final int clauseId = ASTBuilder.addClause(this.edg, parentId, info);
		final Branch branch = this.branches.push(new Branch(clauseId, NodeInfo.Type.Clause, info));

		branch.setWhere(Where.Parameters);
		this.processElements(parameters);
		branch.setWhere(Where.Guard);
		if (guard != null)
			this.processElement(guard, 1, 1);
		branch.setWhere(Where.Body);
		if (body != null)
			this.processElement(body, 1, 1);
		this.branches.pop();
	}
	protected <R, S, T> void addClause(Iterable<R> parameters, S guard, Iterable<T> expressions, LDASTNodeInfo info)
	{
		final R[] parameters0 = this.getArray(parameters);
		final T[] expressions0 = this.getArray(expressions);
		this.addClause(parameters0, guard, expressions0, info);
	}
	protected <R, S, T> void addClause(R[] parameters, S guard, T[] expressions, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final int clauseId = ASTBuilder.addClause(this.edg, parentId, info);
		final Branch branch = this.branches.push(new Branch(clauseId, NodeInfo.Type.Clause, info));

		branch.setWhere(Where.Parameters);
		this.processElements(parameters);
		branch.setWhere(Where.Guard);
		if (guard != null)
			this.processElement(guard, 1, 1);
		branch.setWhere(Where.Body);
		this.processElements(expressions);
		this.branches.pop();
	}

	protected void addVariable(String name, boolean declaration, boolean definition, boolean global, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final Where where = parent.getWhere();

		ASTBuilder.addVariable(this.edg, parentId, where, name, declaration, definition, global, info);
	}
	protected void addLiteral(String value, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final Where where = parent.getWhere();

		ASTBuilder.addLiteral(this.edg, parentId, where, value, info);
	}
	protected <R, S> void addEquality(R left, S right, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final Where where = parent.getWhere();
		final int equalityId = ASTBuilder.addEquality(this.edg, parentId, where, info);

		this.branches.push(new Branch(equalityId, NodeInfo.Type.Equality, info));
		this.processElement(left, 1, 2);
		this.processElement(right, 2, 2);
		this.branches.pop();
	}
	protected <R> void addOperation(String operation, Iterable<R> operands, LDASTNodeInfo info)
	{
		final R[] operands0 = this.getArray(operands);
		this.addOperation(operation, operands0, info);
	}
	protected <R> void addOperation(String operation, R[] operands, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final Where where = parent.getWhere();
		final int operationId = ASTBuilder.addOperation(this.edg, parentId, where, operation, info);

		this.branches.push(new Branch(operationId, NodeInfo.Type.Operation, info));
		this.processElements(operands);
		this.branches.pop();
	}
	protected <R> void addDataConstructor(Iterable<R> elements, LDASTNodeInfo info)
	{
		final R[] elements0 = this.getArray(elements);
		this.addDataConstructor(elements0, info);
	}
	protected <R> void addDataConstructor(R[] elements, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final Where where = parent.getWhere();
		final int dataConstructorId = ASTBuilder.addDataConstructor(this.edg, parentId, where, info);

		this.branches.push(new Branch(dataConstructorId, NodeInfo.Type.DataConstructor, info));
		this.processElements(elements);
		this.branches.pop();
	}
	protected <R> void addList(Iterable<R> elements, LDASTNodeInfo info)
	{
		final R[] elements0 = this.getArray(elements);
		this.addList(elements0, info);
	}
	protected <R> void addList(R[] elements, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final Where where = parent.getWhere();
		final int listId = ASTBuilder.addList(this.edg, parentId, where, info);

		this.branches.push(new Branch(listId, NodeInfo.Type.List, info));
		this.processElements(elements);
		this.branches.pop();
	}
	protected <R, S> void addDataConstructorAccess(R dataConstructor, S access, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final Where where = parent.getWhere();
		final int dataConstructorId = ASTBuilder.addDataConstructorAccess(this.edg, parentId, where, info);

		this.branches.push(new Branch(dataConstructorId, NodeInfo.Type.DataConstructorAccess, info));
		this.processElement(dataConstructor, 1, 2);
		this.processElement(access, 2, 2);
		this.branches.pop();
	}
	protected <R> void addBlock(Iterable<R> expressions, LDASTNodeInfo info)
	{
		final R[] expressions0 = this.getArray(expressions);
		this.addBlock(expressions0, info);
	}
	protected <R> void addBlock(R[] expressions, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final Where where = parent.getWhere();
		final int blockId = ASTBuilder.addBlock(this.edg, parentId, where, info);

		this.branches.push(new Branch(blockId, NodeInfo.Type.Block, info));
		this.processElements(expressions);
		this.branches.pop();
	}
	protected <R, S, T> void addIf(R condition, Iterable<S> thenExpressions, Iterable<T> elseExpressions, LDASTNodeInfo info)
	{
		final S[] thenExpressions0 = this.getArray(thenExpressions);
		final T[] elseExpressions0 = this.getArray(elseExpressions);
		this.addIf(condition, thenExpressions0, elseExpressions0, info);
	}
	protected <R, S, T> void addIf(R condition, S[] thenExpressions, T[] elseExpressions, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final Where where = parent.getWhere();
		final int ifId = ASTBuilder.addIf(this.edg, parentId, where, info);
		final Branch branch = this.branches.push(new Branch(ifId, NodeInfo.Type.If, info));

		branch.setWhere(Where.Condition);
		this.processElement(condition, 1, 1);
		branch.setWhere(Where.Then);
		this.processElements(thenExpressions);
		branch.setWhere(Where.Else);
		this.processElements(elseExpressions);
		this.branches.pop();
	}
	protected <R, S> void addSwitch(R selector, Iterable<S> cases, LDASTNodeInfo info)
	{
		final S[] cases0 = this.getArray(cases);
		this.addSwitch(selector, cases0, info);
	}
	protected <R, S> void addSwitch(R selector, S[] cases, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final Where where = parent.getWhere();
		final int switchId = ASTBuilder.addSwitch(this.edg, parentId, where, info);
		final Branch branch = this.branches.push(new Branch(switchId, NodeInfo.Type.Switch, info));

		branch.setWhere(Where.Selector);
		if (selector != null)
			this.processElement(selector, 1, 1);
		branch.setWhere(Where.Cases);
		this.processElements(cases);
		this.branches.pop();
	}
	protected <R, S, T> void addCase(R selectable, S guard, T body, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final int caseId = ASTBuilder.addCase(this.edg, parentId, info);
		final Branch branch = this.branches.push(new Branch(caseId, NodeInfo.Type.Case, info));

		branch.setWhere(Where.Selectable);
		this.processElement(selectable, 1, 1);
		branch.setWhere(Where.Guard);
		if (guard != null)
			this.processElement(guard, 1, 1);
		branch.setWhere(Where.Body);
		if (body != null)
			this.processElement(body, 1, 1);
		this.branches.pop();
	}
	protected <R, S, T> void addCase(R selectable, S guard, Iterable<T> expressions, LDASTNodeInfo info)
	{
		final T[] expressions0 = this.getArray(expressions);
		this.addCase(selectable, guard, expressions0, info);
	}
	protected <R, S, T> void addCase(R selectable, S guard, T[] expressions, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final int caseId = ASTBuilder.addCase(this.edg, parentId, info);
		final Branch branch = this.branches.push(new Branch(caseId, NodeInfo.Type.Case, info));

		branch.setWhere(Where.Selectable);
		if (selectable != null)
			this.processElement(selectable, 1, 1);
		branch.setWhere(Where.Guard);
		if (guard != null)
			this.processElement(guard, 1, 1);
		branch.setWhere(Where.Body);
		this.processElements(expressions);
		this.branches.pop();
	}
	protected <R> void addDefaultCase(R body, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final int defaultCaseId = ASTBuilder.addDefaultCase(this.edg, parentId, info);
		final Branch branch = this.branches.push(new Branch(defaultCaseId, NodeInfo.Type.DefaultCase, info));

		branch.setWhere(Where.Body);
		if (body != null)
			this.processElement(body, 1, 1);
		this.branches.pop();
	}
	protected <R, S, T> void addDefaultCase(Iterable<T> expressions, LDASTNodeInfo info)
	{
		final T[] expressions0 = this.getArray(expressions);
		this.addDefaultCase(expressions0, info);
	}
	protected <R> void addDefaultCase(R[] expressions, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final int defaultCaseId = ASTBuilder.addDefaultCase(this.edg, parentId, info);
		final Branch branch = this.branches.push(new Branch(defaultCaseId, NodeInfo.Type.DefaultCase, info));

		branch.setWhere(Where.Body);
		this.processElements(expressions);
		this.branches.pop();
	}
	protected <R, S, T> void addCall(R scope, S function, Iterable<T> arguments, LDASTNodeInfo info)
	{
		final T[] arguments0 = this.getArray(arguments);
		this.addCall(scope, function, arguments0, info);
	}
	protected <R, S, T> void addCall(R scope, S function, T[] arguments, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final Where where = parent.getWhere();
		final int callId = ASTBuilder.addCall(this.edg, parentId, where, info);
		final Branch branch = this.branches.push(new Branch(callId, NodeInfo.Type.Call, info));

		branch.setWhere(Where.Scope);
		if (scope != null)
			this.processElement(scope, 1, 2);
		branch.setWhere(Where.Name);
		if (function != null)
			this.processElement(function, 2, 2);
		branch.setWhere(Where.Arguments);
		this.processElements(arguments);
		this.branches.pop();
	}
	protected <R, S> void addListComprehension(Iterable<R> restrictions, S value, LDASTNodeInfo info)
	{
		final R[] restrictions0 = this.getArray(restrictions);
		this.addListComprehension(restrictions0, value, info);
	}
	protected <R, S> void addListComprehension(R[] restrictions, S value, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final Where where = parent.getWhere();
		final int listComprehensionId = ASTBuilder.addListComprehension(this.edg, parentId, where, info);
		final Branch branch = this.branches.push(new Branch(listComprehensionId, NodeInfo.Type.ListComprehension, info));

		branch.setWhere(Where.Restrictions);
		this.processElements(restrictions);
		branch.setWhere(Where.Value);
		if (value != null)
			this.processElement(value, 1, 1);
		this.branches.pop();
	}
	protected <R, S> void addGenerator(R left, S right, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final Where where = parent.getWhere();
		final int generatorId = ASTBuilder.addGenerator(this.edg, parentId, where, info);

		this.branches.push(new Branch(generatorId, NodeInfo.Type.Generator, info));
		if (left != null)
			this.processElement(left, 1, 2);
		if (right != null)
			this.processElement(right, 2, 2);
		this.branches.pop();
	}
	protected <R, S> void addFilter(R filter, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final Where where = parent.getWhere();
		final int filterId = ASTBuilder.addFilter(this.edg, parentId, where, info);

		this.branches.push(new Branch(filterId, NodeInfo.Type.Filter, info));
		this.processElement(filter, 1, 1);
		this.branches.pop();
	}
	protected <R, S, T> void addLoop(R condition, Iterable<S> bodyExpressions, LDASTNodeInfo info)
	{
		final S[] bodyExpressions0 = this.getArray(bodyExpressions);
		this.addLoop(condition, bodyExpressions0, info);
	}
	protected <R, S, T> void addLoop(R condition, S[] bodyExpressions, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final Where where = parent.getWhere();
		final int loopId = ASTBuilder.addLoop(this.edg, parentId, where, info);
		final Branch branch = this.branches.push(new Branch(loopId, NodeInfo.Type.Loop, info));

		branch.setWhere(Where.Condition);
		this.processElement(condition, 1, 1);
		branch.setWhere(Where.Body);
		this.processElements(bodyExpressions);
		this.branches.pop();
	}
	protected <R> void addReturn(R expression, int dstId, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final Where where = parent.getWhere();
		final int returnId = ASTBuilder.addReturn(this.edg, parentId, where, dstId, info);

		this.branches.push(new Branch(returnId, NodeInfo.Type.Return, info));
		if (expression != null)
			this.processElement(expression, 1, 1);
		this.branches.pop();
	}
	protected void addBreak(int dstId, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final Where where = parent.getWhere();

		ASTBuilder.addBreak(this.edg, parentId, where, dstId, info);
	}
	protected void addContinue(int dstId, LDASTNodeInfo info)
	{
		final Branch parent = this.branches.peek();
		final int parentId = parent.getNodeId();
		final Where where = parent.getWhere();

		ASTBuilder.addContinue(this.edg, parentId, where, dstId, info);
	}

	@SuppressWarnings("unchecked")
	private <R> R[] getArray(Iterable<R> iterable)
	{
		final List<R> list = new LinkedList<R>();

		for (R element : iterable)
			list.add(element);

		return (R[]) list.toArray();
	}
	private boolean isPatternZone(List<Branch> ancestors)
	{
		for (int ancestorIndex = ancestors.size() - 1; ancestorIndex >= 0; ancestorIndex--)
		{
			final Branch ancestor = ancestors.get(ancestorIndex);
			final NodeInfo.Type type = ancestor.getNodeType();
			final Where where = ancestor.getWhere();
			final int index = ancestor.getIndex();

			switch (type)
			{
				case Clause:
					if (where == Where.Parameters)
						return true;
					break;
				case Case:
					if (where == Where.Selectable)
						return true;
					break;
				case Equality:
				case Generator:
					if (index == 1)
						return true;
					break;
				default:
					break;
			}
		}

		return false;
	}

	public static class Branch
	{
		private final int nodeId;
		private final NodeInfo.Type nodeType;
		private final LDASTNodeInfo ldNodeInfo;
		private Where where;
		private int index;
		private int length;

		public Branch(int nodeId, NodeInfo.Type nodeType, LDASTNodeInfo ldNodeInfo)
		{
			this.nodeId = nodeId;
			this.nodeType = nodeType;
			this.ldNodeInfo = ldNodeInfo;
		}

		public int getNodeId()
		{
			return this.nodeId;
		}
		public NodeInfo.Type getNodeType()
		{
			return this.nodeType;
		}
		public LDASTNodeInfo getLdASTNodeInfo()
		{
			return this.ldNodeInfo;
		}
		public Where getWhere()
		{
			return this.where;
		}
		public int getIndex()
		{
			return this.index;
		}
		public int getLength()
		{
			return this.length;
		}

		public void setWhere(Where where)
		{
			this.where = where;
		}
		public void setIndex(int index)
		{
			this.index = index;
		}
		public void setLength(int length)
		{
			this.length = length;
		}
	}
}