package edg.edge;

import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import edg.constraint.AsteriskConstraint;
import edg.constraint.Constraints;
import edg.constraint.GrammarConstraint;
import edg.graph.EDG;
import edg.graph.EdgeInfo;
import edg.graph.Grammar;
import edg.graph.Node;
import edg.graph.NodeInfo;
import edg.slicing.ConstrainedAlgorithm;
import edg.slicing.Phase;
import edg.traverser.EDGTraverser;
import edg.work.NodeWork;
import edg.work.Work;
import edg.work.WorkList;

public class SummaryEdgeGenerator extends EdgeGenerator
{
	public SummaryEdgeGenerator(EDG edg)
	{
		super(edg);
	}

	public void generate()
	{
		this.generateExternalSummaryEdges();
		this.generateInternalSummaryEdges();
	}

	/************************************/
	/************* External *************/
	/************************************/
	private void generateExternalSummaryEdges()
	{
		final List<Node> calls = EDGTraverser.getNodes(this.edg, NodeInfo.Type.Call);

		for (Node call : calls)
		{
			final Node calleeNode = EDGTraverser.getChild(call, 0);
			final Node calleeResultNode = EDGTraverser.getChild(calleeNode, 2);
			final List<Node> inputs = EDGTraverser.getInputs(calleeResultNode, EDGTraverser.Direction.Forwards);
			if (!inputs.isEmpty())
				continue;

			final Node callResult = EDGTraverser.getSibling(call, 1);
			final Node arguments = EDGTraverser.getChild(call, 1);
			final List<Node> argumentNodes = EDGTraverser.getChildren(arguments);

			for (Node argumentNode : argumentNodes)
			{
				final Node result = EDGTraverser.getResult(argumentNode);

				if (result != null)
					this.edg.addEdge(result, callResult, 0, new EdgeInfo(EdgeInfo.Type.Summary, AsteriskConstraint.getConstraint()));
			}
		}
	}

	/************************************/
	/************* Internal *************/
	/************************************/
private final Map<String, Integer> ids = new Hashtable<String, Integer>();
	private void generateInternalSummaryEdges()
	{
		final List<Work> initialWorks = this.getInitialWorks();
		final WorkList workList = new WorkList(initialWorks);
		final ConstrainedAlgorithm slicingAlgorithm = new ConstrainedAlgorithm();

		while (workList.hasMore())
		{
			final Work work = workList.next();
			final Node initialNode = work.getInitialNode();
			final Constraints constraints = work.getConstraints();
			boolean isFormalIn = false;

			if (work instanceof NodeWork)
			{
				final NodeWork nodeWork = (NodeWork) work;
				final Node currentNode = nodeWork.getCurrentNode();

				if (isFormalIn = this.isFormalIn(currentNode, initialNode))
				{
					final List<Node> nodesToContinue = this.createSummaryEdges(initialNode, currentNode, constraints);
					this.rependWorks(workList, nodesToContinue);
				}
			}
			workList.done(work);
			if (isFormalIn)
				continue;

			final List<Work> newWorks = slicingAlgorithm.processWork(Phase.SummaryGeneration, work);
			workList.pendAll(newWorks);
		}
	}
	private List<Work> getInitialWorks()
	{
		final List<Work> workList = new LinkedList<Work>();
		final List<Node> routines = EDGTraverser.getNodes(this.edg, NodeInfo.Type.Routine);

		for (Node routine : routines)
		{
			final List<Node> clauses = EDGTraverser.getChildren(routine);

			for (Node clause : clauses)
			{
				final Node clauseResult = EDGTraverser.getChild(clause, 3);
				workList.add(new NodeWork(clauseResult, clauseResult, new Constraints()));
			}
		}

		return workList;
	}
	private boolean isFormalIn(Node node, Node formalOutNode)
	{
		final Node parent = EDGTraverser.getParent(node);
		final NodeInfo.Type nodeType = node.getData().getType();
		final NodeInfo.Type parentType = parent == null ? null : parent.getData().getType();
		final Node grandParent = parent == null ? null : EDGTraverser.getParent(parent);
		final NodeInfo.Type grandParentType = grandParent == null ? null : grandParent.getData().getType();
		if (parent == null)
			return false;
		if (nodeType != NodeInfo.Type.Parameters)
			if (nodeType != NodeInfo.Type.Result || parentType != NodeInfo.Type.Expression || grandParentType != NodeInfo.Type.Parameters)
				return false;

		// The formal in must be related to the formal out
		final Node parameters = EDGTraverser.getAncestor(node, NodeInfo.Type.Parameters);
		final Node clauseResult = EDGTraverser.getSibling(parameters, 3);
		return clauseResult == formalOutNode;
	}
	private List<Node> createSummaryEdges(Node formalOut, Node formalIn, Constraints constraints)
	{
		final Grammar grammar = this.edg.getGrammar();
		final GrammarConstraint grammarConstraint = new GrammarConstraint(grammar, formalIn);
		this.edg.addProduction(grammarConstraint, constraints);

		final List<Node> nodesToContinue = new LinkedList<Node>();
		final List<Node> inputs = EDGTraverser.getInputs(formalIn, EDGTraverser.Direction.Backwards);

		for (Node input : inputs)
		{
			final Node call = EDGTraverser.getAncestor(input, NodeInfo.Type.Call);
			if (call == null)
				continue;

			final Node callResult = EDGTraverser.getSibling(call, 1);
			this.edg.addEdge(input, callResult, 0, new EdgeInfo(EdgeInfo.Type.Summary, grammarConstraint));
			nodesToContinue.add(callResult);
		}

		return nodesToContinue;
	}
	private void rependWorks(WorkList workList, List<Node> nodesToContinue)
	{
		for (Node nodeToContinue : nodesToContinue)
		{
			final String id = nodeToContinue.getData().getId() + "";
			workList.repend(id);
		}
	}
}